/* bt.c - interface for bt program, calls btd and btc */

#include "bt.h"

/*------------------------------------------------------------------------
 * main - interface for BT service
 *------------------------------------------------------------------------
 */
int
main(int argc, char *argv[])
{
	pthread_t	th;
	pthread_attr_t	ta;

	nfiles=0; /* startup value */

 	files = malloc(NUMFILES * sizeof(filedesc));
	memset(files, 0, (NUMFILES * sizeof(filedesc)));

	logaudit("--------------------------------------------------------------------SESSION START\n");
	/*if filestate file exists, load it; otherwise, create a new one*/
	if (access(filestate, F_OK)<0) {
		printf("bt: main, no filestate file found, running updatefiles\n");
		updatefiles();
	} else {
		printf("bt: main, filestate file found, running loadstate\n");
		loadstate();
		printf("bt: main, loadstate complete, running updatefiles\n");
		updatefiles();
	}

	/*switch (argc) {
	case	1:
		break;
	case	2:
		service = argv[1];
		break;
	default:
		errexit("usage: bt [port]\n");
	}*/

	(void) pthread_attr_init(&ta);
	(void) pthread_attr_setdetachstate(&ta, PTHREAD_CREATE_DETACHED);
	(void) pthread_mutex_init(&stats.st_mutex, 0);
	(void) pthread_mutex_init(&filestate_mutex, 0);

	/*printf("bt: main, starting thread for prstats\n");
	if (pthread_create(&th, &ta, (void * (*)(void *))prstats, 0) < 0)
		errexit("pthread_create(prstats): %s\n", strerror(errno));*/

	printf("bt: main, starting thread for btd\n");
	if (pthread_create(&th, &ta, (void * (*)(void *))btd, 0) < 0)
		errexit("pthread_create(btd): %s\n", strerror(errno));

	printf("bt: main, starting thread for btc\n");
	if (pthread_create(&th, &ta, (void * (*)(void *))btc, 0) < 0)
		errexit("pthread_create(btc): %s\n", strerror(errno));

	/* for real, this would be the place to put command-line interface things 
	but for now, we've created the important threads, so wait forever then return 0 */
	printf("bt: main, sleeping forever\n");

	while (1) {
		interface();
	//	sleep(13371337);
	}

	return 0;

}

int interface(void) {

	char buf[BUFSIZE];
	char fname[BUFSIZE];
	char c;
	int i;
	int status;
	pthread_t	th;
	pthread_attr_t	ta;

	(void) pthread_attr_init(&ta);
	(void) pthread_attr_setdetachstate(&ta, PTHREAD_CREATE_DETACHED);

	memset(buf, 0, BUFSIZE);
	memset(fname, 0, BUFSIZE);

	printf("Enter command (h for help): ");
	fgets(buf, BUFSIZE, stdin);
	c=buf[0];
	
	switch (c) {
		case 'h':
			printf("Available commands:\n");
			printf("\th: (h)elp\n");
			printf("\tf: print loaded (f)iles\n");
			printf("\tl: (l)oad new file\n");
			printf("\tn: create (n)ew metafile\n");
			printf("\tp: (p)ause or un(p)ause a file\n");
			printf("\tq: (q)uit\n");
			printf("\ts: print (s)tatistics\n");
			printf("\tx: stop all data transfer for a file\n");			
			break;
		case 'f':
			printf("Currently loaded files: \n");
			prfiles();
			break;
		case 'l':
			printf("Enter the name of a new metafile to load (extension .bt): ");
			fgets(fname, BUFSIZE, stdin);
			//strip trailing newline
			fname[strlen(fname)-1]=0;
			if (access(fname, F_OK)<0) {
				printf("File %s not found\n", fname);
			} else {
				status=loadmetafile(fname);
				if (status<0) {
					printf("Error loading %s. Check %s for details.\n", fname, auditlog);
				} else if (status>0) {
					printf("File %s already loaded.\n", fname);
				} else {
					printf("File %s loaded.\nDownload missing pieces of %.*s? (y/n): ", fname, FILENAMESIZ, files[nfiles-1].filename);
					fgets(buf, BUFSIZE, stdin);	
					c=buf[0];
					if (c=='y') {
						//This creates a possible race condition if any other threads change nfiles.
						printf("Starting download of %.*s\n", FILENAMESIZ, files[nfiles-1].filename);
						if (pthread_create(&th, &ta, (void * (*)(void *))getfile, (void *)&files[nfiles-1]) < 0) {
							logaudit("interface: pthread_create: %s, exiting\n", strerror(errno));
							return 0;
						}
					} else {
						printf("Not downloading. Any missing pieces of %.*s will be downloaded next session.\n", FILENAMESIZ, files[nfiles-1].filename);
					}	
				}
			}
			break;
		case 'n':
			printf("Enter the name of the file to publish: ");
			fgets(fname, BUFSIZE, stdin);
			//strip trailing newline
			fname[strlen(fname)-1]=0;
			if (access(fname, F_OK)<0) {
				printf("File %s not found\n", fname);
				return 0;
			} else {
				createmetafile(fname);
				printf("Metafile for %s created. Begin hosting? (y/n): ", fname);
				fgets(buf, BUFSIZE, stdin);	
				c=buf[0];
				if (c=='y') {	
					strcat(fname, METAFILEEXT);
					printf("Loading %s.\n", fname);
					status=loadmetafile(fname);
					if (status<0) {
						printf("Error loading %s. Check %s for details.\n", fname, auditlog);
					} else {
						printf("Now hosting. Distribute %s so that others can download from you.\n", fname);
					}
				} else {
					printf("Hosting %s will begin next session.\n", fname);
				}			
			}			
			break;
		case 'p':
			printf("Currently loaded files:\n");
			prfiles();
			printf("Enter the number of the file to pause or unpause (c to cancel): ");
			fgets(buf, BUFSIZE, stdin);
			if (buf[0]=='c') {
				return 0;
			}
			i=atoi(&buf[0]);
			(void) pthread_mutex_lock(&files[i].filedesc_mutex);
			files[i].sessionok=!files[i].sessionok;
			(void) pthread_mutex_unlock(&files[i].filedesc_mutex);
			if (files[i].sessionok==TRANSFER_OK) {
				printf("Starting download of %.*s\n", FILENAMESIZ, files[i].filename);
				if (pthread_create(&th, &ta, (void * (*)(void *))getfile, (void *)&files[i]) < 0) {
					logaudit("interface: pthread_create: %s, exiting\n", strerror(errno));
					return 0;
				}				
			}
			break;			
		case 'q':
			printf("Saving state and exiting.\n");
			savestate();
			exit(0);
			break;
		case 's':
			prstats();
			break;
		case 'x':
			printf("Currently loaded files:\n");
			prfiles();
			printf("This operation permanently stops all transfer of the selected file.\n");
			printf("Enter the number of the file to stop transferring (c to cancel): ");
			fgets(buf, BUFSIZE, stdin);
			if (buf[0]=='c') {
				return 0;
			}
			i=atoi(&buf[0]);
			(void) pthread_mutex_lock(&files[i].filedesc_mutex);
			files[i].transferok=TRANSFER_STOP;
			files[i].sessionok=files[i].transferok;
			(void) pthread_mutex_unlock(&files[i].filedesc_mutex);	
			savestate();	
			printf("All transfer of file %s stopped permanently.\n", files[i].filename);	
			break;
		default: 
			printf("Command %c not recognized\n", c);
			break;
	}
	return 0;
}	
