#include <netdb.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <dirent.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/errno.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <arpa/inet.h>


#ifndef _SVID_SOURCE
#define _SVID_SOURCE
#endif // _SVID_SOURCE

#define	QLEN		  32	/* maximum connection queue length	*/
#define	BUFSIZE		4096

#define TRANSFER_OK	1
#define TRANSFER_STOP	0

#define	INTERVAL	5	/* secs */

#define BTNAME_MAX	256
#define MAXADDRLEN	128

#define FILENAMESIZ	128
#define MSGTYPESIZ	4
#define CHUNKNOSIZ	8
#define CHUNKLENGTHSIZ	8
#define MSGHEADERSIZ 	FILENAMESIZ+MSGTYPESIZ+CHUNKNOSIZ+CHUNKLENGTHSIZ

/* message types must be 4 bytes long*/ 
#define CHUNKMSG 	"data"
#define ADDRMSG		"addr" 
#define BADMSGMSG	"what"
#define FILENOTFOUNDMSG "nope"

#define ADDRCHUNKNO	-1

#define METAFILEEXT	".bt"

#define CHUNKSIZE 	(8 * 1024) /* chunksize for all files is 8k, in bytes */

#define NUMFILES 	8 /*maximum number of files this program can handle */

#define MAXCLIENTCONNS	20 /*maximum number of connections the client is allowed to make */

#define BTPORT		"1337"

#define errorlog 	"errorlog.txt"

#define filestate	"filestate.txt"

#define auditlog 	"auditlog.txt"

#pragma pack (1)
typedef struct {
	char filename[FILENAMESIZ];
	char msgtype[MSGTYPESIZ];
	char chunkno[CHUNKNOSIZ];
	char chunksize[CHUNKLENGTHSIZ];
	char data[CHUNKSIZE];
} btmsg;
#pragma pack ()

typedef struct {
	char filename[FILENAMESIZ+1];
	char msgtype[MSGTYPESIZ+1];
	char chunkno[CHUNKNOSIZ];
	char chunksize[CHUNKLENGTHSIZ];
	char data[CHUNKSIZE];
	char host[MAXADDRLEN+1];
	int addrlen;
	int filesindex;
	int checksum;
} btintmsg;

typedef struct {
	int chunkno;
	int checksum;
	int want;
	int have;
} chunk;

typedef struct {
	char host[MAXADDRLEN];
	int addrlen;
} addrs;

typedef struct {
	pthread_mutex_t filedesc_mutex;
	int filesize;
	char filename[FILENAMESIZ+1];
	int chunksize;	
	/*char path;*/
	int nchunks;
	chunk* chunks;
	pthread_mutex_t addresses_mutex;
	addrs* addresses;
	int naddrs;
	int naddrsused;
	int filesindex;
	int transferok;
	int sessionok; //should not be saved.
} filedesc;

struct statsobj {
	pthread_mutex_t	st_mutex;
	unsigned int	st_concount;
	unsigned int	st_contotal;
	unsigned long	st_contime;
	unsigned long	st_bytecount;
	unsigned long	nclientconns;
};


/*initializing some globals*/
extern filedesc* files;
extern int nfiles;
extern pthread_mutex_t filestate_mutex;
extern struct statsobj stats;

/*function prototypes*/
extern void prstats(void);
extern int btdresp(void* insock);
extern int errexit(const char *format, ...);
extern int passiveTCP(const char *service, int qlen);
extern int btc(void);
extern int btd(void);
extern int getfile(void* file);
extern int getchunk(void* servmsg);
extern int getaddrs (btintmsg servmsg);
extern int removeeaddrs (btintmsg clientmsg, int sock);
extern int createmetafile(char* filename);
extern int removeeaddrs (btintmsg clientmsg, int sock);
extern int updateaddrs (btintmsg clientmsg, int sock);
extern int addaddrslist (btintmsg clientmsg, addrs* addrlist, int naddrs);
extern int sendchunk (int sock, btintmsg clientmsg);
extern int sendaddrs (int sock, btintmsg clientmsg);
extern int sendbadmsg (int sock, btintmsg clientmsg);
extern int sendnotfound (int sock, btintmsg clientmsg);
extern int loadstate(void);
extern int updatefiles(void);
extern int savestate(void);
extern void prstats(void);
extern int logerror (int sock, char* request, int length);
extern int infiletable (char *filename);
extern int recvheader (btintmsg *header, int sock);
extern int logaudit(const char *format, ...);
extern int interface(void);
extern int loadmetafile (char *fname);
extern int whereinfiletable (char *filename);
extern int prfiles (void);

