/* updateaddrs, removeaddrs - add or remove addresses from the list of addresses associated with a filedesc */

#include "bt.h"

/*------------------------------------------------------------------------
 * removeeaddrs -if we get a bogus reply from a server, remove its address from the list and save
 *------------------------------------------------------------------------
 */

int removeaddrs (btintmsg clientmsg) {

	int i;

	logaudit("removeaddrs: IP address to remove (clientmsg.host) is: %.*s, addrlen is %d\n", clientmsg.addrlen, clientmsg.host, clientmsg.addrlen);

	(void) pthread_mutex_lock(&files[clientmsg.filesindex].addresses_mutex);
	for (i=0; i<files[clientmsg.filesindex].naddrsused; i++) {
		logaudit("removeaddrs: files[clientmsg.filesindex].addresses[%d].host is %s\n", i, files[clientmsg.filesindex].addresses[i].host);
		if (!(strncmp(files[clientmsg.filesindex].addresses[i].host, clientmsg.host, clientmsg.addrlen))) {
			logaudit("removeaddrs: found address on the list\n");
			break;
		}
	}

	if (i==files[clientmsg.filesindex].naddrsused) {
		logaudit("removeaddrs: address %.*s not found, exiting.\n", clientmsg.addrlen, clientmsg.host);
		(void) pthread_mutex_unlock(&files[clientmsg.filesindex].addresses_mutex);
		return 0;
	} else if (files[clientmsg.filesindex].naddrsused==1) {
		logaudit("removeaddrs: address %.*s is the only address on the list. Not removing.\n", clientmsg.addrlen, clientmsg.host);
		logaudit("removeaddrs: setting sessionok to TRANSFER_STOP so that getfile can give up.\n");
		(void) pthread_mutex_lock(&files[clientmsg.filesindex].filedesc_mutex);
		files[clientmsg.filesindex].sessionok=TRANSFER_STOP;
		(void) pthread_mutex_unlock(&files[clientmsg.filesindex].filedesc_mutex);
		logaudit("removeaddrs: saving state\n");
	} else {
		logaudit("removing address %.*s from the list in element %d\n", clientmsg.addrlen, clientmsg.host, i);
		memset(files[clientmsg.filesindex].addresses[i].host, 0, MAXADDRLEN);
		files[clientmsg.filesindex].addresses[i].addrlen=0;
		files[clientmsg.filesindex].naddrsused--;
		logaudit("removeaddrs: saving state\n");

	}
	(void) pthread_mutex_unlock(&files[clientmsg.filesindex].addresses_mutex);
	savestate();
	return 0;
}	

/*------------------------------------------------------------------------
 * updateaddrs - when we send a chunk, add the address to the list
 *------------------------------------------------------------------------
 */

int updateaddrs (btintmsg clientmsg, int sock) {

	struct sockaddr_in cli_addr;
	socklen_t sin_len;
	int i;

	sin_len = sizeof (cli_addr);

	if (getpeername (sock, (struct sockaddr *) &cli_addr, &sin_len) < 0) {
		logaudit("updateaddrs: chunk %d of %s, getpeername: %s, not updating addresses\n", clientmsg.chunkno, clientmsg.filename, strerror(errno));
	}

	logaudit("updateaddrs: client's IP address (inet_ntoa(cli_addr.sin_addr)) is: %s\n", inet_ntoa(cli_addr.sin_addr));

	for (i=0; i<files[clientmsg.filesindex].naddrsused; i++) {
		logaudit("updateaddrs: files[clientmsg.filesindex].addresses[%d].host is %s\n", i, files[clientmsg.filesindex].addresses[i].host);
		if (!(strcmp(files[clientmsg.filesindex].addresses[i].host, inet_ntoa(cli_addr.sin_addr)))) {
			logaudit("updateaddrs: found address on the list, not adding\n");
			break;
		}
	}

	if (i==files[clientmsg.filesindex].naddrsused) {
		logaudit("updateaddrs: address %s not found, adding\n", inet_ntoa(cli_addr.sin_addr));
		(void) pthread_mutex_lock(&files[clientmsg.filesindex].addresses_mutex);
		files[clientmsg.filesindex].naddrsused++;
		if (files[clientmsg.filesindex].naddrsused >= files[clientmsg.filesindex].naddrs) {
			files[clientmsg.filesindex].naddrs+=10;
			files[nfiles].addresses = realloc(files[nfiles].addresses, files[nfiles].naddrs * sizeof(addrs));
		}
		logaudit("updateaddrs: copying address %s to the list in element %d\n", inet_ntoa(cli_addr.sin_addr), files[clientmsg.filesindex].naddrsused);
		strcpy(files[clientmsg.filesindex].addresses[files[clientmsg.filesindex].naddrsused-1].host, inet_ntoa(cli_addr.sin_addr));
		logaudit("updateaddrs: setting address length to %d\n", MAXADDRLEN);
		files[clientmsg.filesindex].addresses[files[clientmsg.filesindex].naddrsused-1].addrlen=MAXADDRLEN;
		logaudit("updateaddrs: files[clientmsg.filesindex].addresses[files[clientmsg.filesindex].naddrsused-1].host is %s\n", files[clientmsg.filesindex].addresses[files[clientmsg.filesindex].naddrsused-1].host);
		(void) pthread_mutex_unlock(&files[clientmsg.filesindex].addresses_mutex);

		logaudit("updateaddrs: saving state\n");
		savestate();
	}
	return 0;
}

/*------------------------------------------------------------------------
 * addaddrslist - add a received list of addresses to the address list
 *------------------------------------------------------------------------
 */

int addaddrslist (btintmsg clientmsg, addrs* addrlist, int naddrs) {

	int i;
	int j;

	for (j=0; j<naddrs; j++) {

		logaudit("addaddrslist: address to check is: %.*s\n", addrlist[j].addrlen, addrlist[j].host);

		for (i=0; i<files[clientmsg.filesindex].naddrsused; i++) {
			logaudit("addaddrslist: files[clientmsg.filesindex].addresses[%d].host is %s\n", i, files[clientmsg.filesindex].addresses[i].host);
			if (!(strncmp(files[clientmsg.filesindex].addresses[i].host, addrlist[j].host, addrlist[j].addrlen))) {
				logaudit("addaddrslist: found address on the list, not adding\n");
				break;
			}
		}
	
		if (i==files[clientmsg.filesindex].naddrsused) {
			logaudit("addaddrslist: address %.*s not found, adding\n", addrlist[j].addrlen, addrlist[j].host);
			(void) pthread_mutex_lock(&files[clientmsg.filesindex].addresses_mutex);
			files[clientmsg.filesindex].naddrsused++;
			if (files[clientmsg.filesindex].naddrsused >= files[clientmsg.filesindex].naddrs) {
				files[clientmsg.filesindex].naddrs+=10;
				files[nfiles].addresses = realloc(files[nfiles].addresses, files[nfiles].naddrs * sizeof	(addrs));
			}
			logaudit("addaddrslist: copying address %.*s to the list in element %d\n", addrlist[j].addrlen, addrlist[j].host, files	[clientmsg.filesindex].naddrsused);
			strncpy(files[clientmsg.filesindex].addresses[files[clientmsg.filesindex].naddrsused-1].host, 	addrlist[j].host, addrlist[j].addrlen);
			logaudit("addaddrslist: setting address length to %d\n", addrlist[j].addrlen);
			files[clientmsg.filesindex].addresses[files[clientmsg.filesindex].naddrsused-1].addrlen=addrlist[j].addrlen;
			logaudit("addaddrslist: files[clientmsg.filesindex].addresses[files[clientmsg.filesindex].naddrsused-1].host is %s\n", files[clientmsg.filesindex].addresses[files[clientmsg.filesindex].naddrsused-1].host);
			(void) pthread_mutex_unlock(&files[clientmsg.filesindex].addresses_mutex);
	
			logaudit("addaddrslist: saving state\n");
			savestate();
		}
	}
	return 0;
}

