/* btc.c - client for bt protocol */

#include "bt.h"

/*------------------------------------------------------------------------
 * btc - Concurrent TCP client for BT service 
 *------------------------------------------------------------------------
 */

int btc(void) {
	
	pthread_t	th;
	pthread_attr_t	ta;
	(void) pthread_attr_init(&ta);
	(void) pthread_attr_setdetachstate(&ta, PTHREAD_CREATE_DETACHED);

	(void) pthread_mutex_lock(&stats.st_mutex);
	stats.nclientconns=0;
	(void) pthread_mutex_unlock(&stats.st_mutex);

	int i;
	int j;
	int getfilethreads=0;

	logaudit("btc: looking for files with missing pieces\n");
	for (i=0; i<nfiles; i++) {
		if (files[i].sessionok==TRANSFER_STOP) {
			logaudit("btc: %s is not available, not starting a thread\n", files[i].filename);
		} else {
			for (j=0; j<files[i].nchunks; j++) {
				if (files[i].chunks[j].want==1) {
					logaudit("btc: started a thread to get %s\n", files[i].filename);
					printf("btc: started a thread to get %s\n", files[i].filename);
					getfilethreads++;
					if (pthread_create(&th, &ta, (void * (*)(void *))getfile, (void *)&files[i]) < 0) {
						logaudit("btc: pthread_create: %s, exiting\n", strerror(errno));
						return 0;
					}
					break;
				}
			}
		}
	}

	logaudit("btc: started %d getfile threads, my work is done.\n", getfilethreads);
	return 0;
}

/*------------------------------------------------------------------------
 * getfile - gets outstanding pieces of one file 
 *------------------------------------------------------------------------
 */

int getfile (void* infile) {
	
	btintmsg servmsg;
	int i;
	pthread_t	th;
	pthread_attr_t	ta;
	filedesc *file = infile;
	int nmissing=file->nchunks;
	char temp[CHUNKNOSIZ+1];
	char temp2[CHUNKLENGTHSIZ+1];
	btintmsg* servmsgs[MAXCLIENTCONNS];
	int counter=0;

	(void) pthread_attr_init(&ta);
	(void) pthread_attr_setdetachstate(&ta, PTHREAD_CREATE_DETACHED);

	for (counter=0; counter<MAXCLIENTCONNS; counter++) {
		servmsgs[counter] = (btintmsg *) malloc(sizeof(btintmsg));
		//printf("proof we can access *socks[%d]: %d\n", counter, *socks[counter]);
	}

	logaudit("getfile: looking for missing pieces of %s, maximum %d\n", file->filename, nmissing);
	while ((nmissing>0) && (files[file->filesindex].sessionok==TRANSFER_OK)){
		counter=(counter+1)%MAXCLIENTCONNS;
		for (i=0; i<file->nchunks; i++) {
			if (files[file->filesindex].sessionok==TRANSFER_STOP) {
				logaudit("getfile: %s is no longer okay to transfer, no new threads this session\n", file->filename);
				break;
			}
			logaudit("getfile: investigating chunk %d of %s\n", i, file->filename);
			if((file->chunks[i].want==0) && (file->chunks[i].have==1)) {
				logaudit("getfile: have chunk %d of %s\n", i, file->filename);
				nmissing--;
			} else {
				logaudit("getfile: missing chunk %d of %s\n", i, file->filename);
				while ((stats.nclientconns>=MAXCLIENTCONNS) || ((stats.nclientconns>=file->nchunks) && (stats.nclientconns>0))) {
					logaudit("getfile: too many client connections (%d of %d), waiting\n", stats.nclientconns, MAXCLIENTCONNS);
					sleep(INTERVAL);
				}
				memcpy(servmsg.filename, file->filename, FILENAMESIZ);
				servmsg.filename[FILENAMESIZ]=0;
				//printf("getfile: servmsg.filename is %s\n", servmsg.filename);

				memset(&servmsg.chunkno, 0, CHUNKNOSIZ);
				sprintf(temp, "%d", i);
				memcpy(servmsg.chunkno, temp, CHUNKNOSIZ);
				//printf("getfile: servmsg.chunkno is %.8s\n", servmsg.chunkno);

				memset(&servmsg.chunksize, 0, CHUNKLENGTHSIZ);
				sprintf(temp2, "%d", file->chunksize);
				memcpy(servmsg.chunksize, temp2, CHUNKLENGTHSIZ);
				//printf("getfile: servmsg.chunksize is %.8s\n", servmsg.chunksize);

				//printf("getfile: i%%file->naddrsused is %d\n", i%file->naddrsused);
				//printf("getfile: file->addresses[%d] is %.*s\n", i%file->naddrsused, MAXADDRLEN, file->addresses[(i%file->naddrsused)].host);
				servmsg.addrlen=file->addresses[(i%file->naddrsused)].addrlen;
				//printf("getfile: servmsg.addrlen is %d\n", servmsg.addrlen);

				memcpy(servmsg.host, file->addresses[(i%file->naddrsused)].host, MAXADDRLEN);
				servmsg.host[servmsg.addrlen]=0;
				//printf("getfile: servmsg.host is %s\n", servmsg.host);

				servmsg.checksum=file->chunks[i].checksum;
				//printf("getfile: servmsg.checksum is %d\n", servmsg.checksum);
				servmsg.filesindex=file->filesindex;
				//printf("getfile: servmsg.filesindex is %d\n", servmsg.filesindex);
				
				logaudit("getfile: chunk %i of %s is missing, starting thread\n", i, file->filename);
				(void) pthread_mutex_lock(&stats.st_mutex);
				stats.st_concount++;
				stats.nclientconns++;
				(void) pthread_mutex_unlock(&stats.st_mutex);

				*servmsgs[counter]=servmsg;
				if (pthread_create(&th, &ta, (void * (*)(void *))getchunk,(void *)servmsgs[counter]) < 0) {
					logaudit("getfile: pthread_create: %s, exiting\n", strerror(errno));
					return 0;
				}
			}
		}

		if (nmissing==0) {
			break;
		} else {
			logaudit("getfile: file->nchunks == %d\n", file->nchunks);
			logaudit("getfile: files[%d].nchunks == %d\n", file->filesindex, files[file->filesindex].nchunks);
			nmissing=file->nchunks; /*reset*/
		}
	}
	if (files[file->filesindex].sessionok==TRANSFER_STOP) {
		logaudit("getfile: transfer of %s not permitted, no new threads this session.\n", file->filename);
	} else {
		logaudit("getfile: successfully received all %d chunks of %s.\n", file->nchunks, file->filename);
		printf("getfile: successfully received all %d chunks of %s.\n", file->nchunks, file->filename);
	}	
	return 0;
}

/*------------------------------------------------------------------------
 * getchunk - gets the specified chunk from the specified server 
 *------------------------------------------------------------------------
 */
	
int getchunk (void* inmsg) {

	btintmsg servmsg = * (btintmsg *) inmsg;
	FILE* fp;
	int sock;
	int bytescopied=0;
	int count;
	char buf[BUFSIZE];
	int chunk2write=atoi(servmsg.chunkno);
	char servreply[BUFSIZ];
	char databuf[BUFSIZ];
	char servdata[CHUNKSIZE];
	btintmsg reply;
	btmsg request;
	int checksum;
	int byteswritten;
	long offset;
	int i;
	long filepos;
	time_t start;
	
	memcpy(request.filename,servmsg.filename,FILENAMESIZ);
	memcpy(request.msgtype, CHUNKMSG, MSGTYPESIZ);
	memcpy(request.chunkno, servmsg.chunkno, CHUNKNOSIZ);
	memcpy(request.chunksize, servmsg.chunksize, CHUNKLENGTHSIZ);

	start = time(0);

	logaudit("getchunk: connecting to %s and requesting chunk %.*s of %s\n", servmsg.host, CHUNKNOSIZ, servmsg.chunkno, servmsg.filename);

	if ((sock=connectTCP(servmsg.host, BTPORT))<0) {
		logaudit("getchunk(connectTCP): failed to connect, removing address %s\n", servmsg.host);
		removeaddrs(servmsg);
		pthread_mutex_lock(&stats.st_mutex);
		stats.st_concount--;
		stats.nclientconns--;
		pthread_mutex_unlock(&stats.st_mutex);
		return 0;
	} else {
		if (write(sock, &request, sizeof request) < 0) {
			logaudit("getchunk: write: %s\n", strerror(errno));
			pthread_mutex_lock(&stats.st_mutex);
			stats.st_concount--;
			stats.nclientconns--;
			pthread_mutex_unlock(&stats.st_mutex);
			return 0;
		}
	}

	recvheader(&reply, sock);

	memcpy(reply.host, servmsg.host, servmsg.addrlen);
	reply.addrlen=servmsg.addrlen;
	reply.filesindex=servmsg.filesindex;
	//printf("getchunk: reply.host is %.*s, reply.addrlen is %d.\n", reply.addrlen, reply.host, reply.addrlen);

	if (strcmp(servmsg.filename, reply.filename)!=0) {
		logaudit("getchunk: requested %s, server sent %s. removing address.\n", servmsg.filename, reply.filename);
		removeaddrs(reply);
		pthread_mutex_lock(&stats.st_mutex);
		stats.st_concount--;
		stats.st_contotal++;
		stats.nclientconns--;
		stats.st_contime += time(0) - start;
		pthread_mutex_unlock(&stats.st_mutex);
		return 0;	
	}

	if (strncmp(reply.msgtype, FILENOTFOUNDMSG, MSGTYPESIZ)==0) {
		logaudit("getchunk: server said file %s not found, remove address and exit\n", servmsg.filename);
		removeaddrs(reply);
		pthread_mutex_lock(&stats.st_mutex);
		stats.st_concount--;
		stats.nclientconns--;
		stats.st_contotal++;
		stats.st_contime += time(0) - start;
		pthread_mutex_unlock(&stats.st_mutex);
		return 0;
	}
	
	if (strncmp(request.chunkno, reply.chunkno, CHUNKNOSIZ)!=0) {
		/*server sent different chunk than the one we requested */
		if (files[servmsg.filesindex].chunks[atoi(reply.chunkno)].have==1) {
			logaudit("getchunk: server sent a chunk of %s we already have\n", servmsg.filename);
			pthread_mutex_lock(&stats.st_mutex);
			stats.st_concount--;
			stats.nclientconns--;
			stats.st_contotal++;
			stats.st_contime += time(0) - start;
			pthread_mutex_unlock(&stats.st_mutex);
			logaudit("getchunk: requesting addresses for %s from server\n", servmsg.filename);
			(void) pthread_mutex_lock(&stats.st_mutex);
			stats.st_concount++;
			stats.nclientconns++;	
			(void) pthread_mutex_unlock(&stats.st_mutex); 
			getaddrs(reply);
			return 0;
		} else {
			chunk2write=atoi(reply.chunkno);
		}
	}

	/* read in the data */
	bytescopied=0;
	do {
		count = read(sock, databuf, atoi(reply.chunksize));
		logaudit("getchunk: reading data, count is: %d\n", count);
		if (count < 0) {
			logaudit("getchunk read msgdata: %s\n", strerror(errno));
			(void) pthread_mutex_lock(&stats.st_mutex);
			stats.st_concount--;
			stats.nclientconns--;
			(void) pthread_mutex_unlock(&stats.st_mutex); 
			return 0;
		}
		memcpy (&servdata[bytescopied], databuf, count);	
		bytescopied+=count;	
		(void) pthread_mutex_lock(&stats.st_mutex);
		stats.st_bytecount += count;
		(void) pthread_mutex_unlock(&stats.st_mutex); 
	} while (bytescopied < atoi(reply.chunksize));

	/* check chunk checksum */
	checksum=0;
	for (i=0; i<bytescopied; i++) {
		checksum+=servdata[i];
	}
	//printf("getchunk: checksum sum is %d\n", checksum);
	checksum=checksum%256;

	//check against the chunk we actually got.
	logaudit("getchunk: calculated checksum is %d, known checksum is %d\n", checksum, files[servmsg.filesindex].chunks[chunk2write].checksum);
	if (checksum!=files[servmsg.filesindex].chunks[chunk2write].checksum) {
		logaudit("getchunk: server sent chunk %d of %s with non-matching checksum, removing address\n", chunk2write, servmsg.filename);
		removeaddrs(reply);
		pthread_mutex_lock(&stats.st_mutex);
		stats.st_concount--;
		stats.nclientconns--;
		stats.st_contotal++;
		stats.st_contime += time(0) - start;
		pthread_mutex_unlock(&stats.st_mutex);
		return 0;
	}

	close(sock); /*the server already closed, this is for completeness*/

	pthread_mutex_lock(&stats.st_mutex);
	stats.nclientconns--;
	stats.st_concount--;
	stats.st_contotal++;
	stats.st_contime += time(0) - start;
	pthread_mutex_unlock(&stats.st_mutex);

	(void) pthread_mutex_lock(&files[servmsg.filesindex].filedesc_mutex);

	/*check for chunk we already have. If another thread got it, abort*/
	if ((files[servmsg.filesindex].chunks[chunk2write].want==0) && (files[servmsg.filesindex].chunks[chunk2write].have==1)) {
		logaudit("getchunk: beaten to saving chunk %d of %s, discarding and exiting\n", chunk2write, reply.filename);
		(void) pthread_mutex_unlock(&files[servmsg.filesindex].filedesc_mutex);
		return 0;
	}

	(void) pthread_mutex_unlock(&files[servmsg.filesindex].filedesc_mutex);

	/* we have the chunk and it's the right data, now save it to disk */
	offset=chunk2write*files[servmsg.filesindex].chunksize;

	/*check if the file exists already, and if not then create it*/
	if (access(reply.filename, F_OK)<0) {
		logaudit("getchunk: file %s does not exist, creating\n", reply.filename);
		fp=fopen(reply.filename, "w");
	} else {
		logaudit("getchunk: file %s exists, opening in mode r+\n", reply.filename);
		fp=fopen(reply.filename, "r+");
	}

	/*regardless of fopen mode, this fseek should succeed*/
	fseek(fp, offset, SEEK_SET);
	filepos=ftell(fp);

	logaudit("getchunk: position in file %s is currently %d, it should be %d\n", reply.filename, filepos, chunk2write*files[servmsg.filesindex].chunksize);
	
	if((byteswritten=fwrite(servdata, 1, atoi(reply.chunksize), fp))<=0) {
		logaudit("getchunk: %s\n", strerror(errno));
		return 0; 
	}

	logaudit("getchunk: wrote %d bytes to %s\n", byteswritten, reply.filename);
	fclose(fp);
	
	logaudit("getchunk: successfully received chunk %d of %s, updating filetable\n", chunk2write, reply.filename);

	(void) pthread_mutex_lock(&files[servmsg.filesindex].filedesc_mutex);
	files[servmsg.filesindex].chunks[chunk2write].want=0;
	files[servmsg.filesindex].chunks[chunk2write].have=1;

	(void) pthread_mutex_unlock(&files[servmsg.filesindex].filedesc_mutex);
	logaudit("getchunk: saving state\n");
	savestate();

	if (chunk2write==0) {
		(void) pthread_mutex_lock(&stats.st_mutex);
		stats.st_concount++;
		stats.nclientconns++;	
		(void) pthread_mutex_unlock(&stats.st_mutex); 
		getaddrs(reply);
	}

	return 0;

}		

int getaddrs (btintmsg servmsg) {

	btmsg request;
	btintmsg reply;
	int sock;
	addrs* addresses;
	int bytescopied;
	int count;
	char databuf[BUFSIZ];
	char servdata[CHUNKSIZE];
	int i;
	time_t start;
	
	memcpy(request.filename,servmsg.filename,FILENAMESIZ);
	memcpy(request.msgtype, ADDRMSG, MSGTYPESIZ);
	//these two lines will be discarded by the server
	memcpy(request.chunkno, servmsg.chunkno, CHUNKNOSIZ);
	memcpy(request.chunksize, servmsg.chunksize, CHUNKLENGTHSIZ);	

	logaudit("getaddrs: connecting to %s and requesting chunk %.*s of %s\n", servmsg.host, CHUNKNOSIZ, servmsg.chunkno, servmsg.filename);

	start=time(0);
	if ((sock=connectTCP(servmsg.host, BTPORT))<0) {
		logaudit("getaddrs(connectTCP): %s\n", strerror(errno));
		(void) pthread_mutex_lock(&stats.st_mutex);
		stats.st_concount--;
		stats.nclientconns--;
		(void) pthread_mutex_unlock(&stats.st_mutex); 	
		return 0;	
	} else {
		if (write(sock, &request, sizeof request) < 0) {
			logaudit ("getaddrs: %s\n", strerror(errno));
			(void) pthread_mutex_lock(&stats.st_mutex);
			stats.st_concount--;
			stats.nclientconns--;
			(void) pthread_mutex_unlock(&stats.st_mutex); 
			return 0;
		}
	}

	recvheader(&reply, sock);
		
	memcpy(reply.host, servmsg.host, servmsg.addrlen);
	reply.addrlen=servmsg.addrlen;
	reply.filesindex=servmsg.filesindex;
	//printf("getchunk: reply.host is %.*s, reply.addrlen is %d.\n", reply.addrlen, reply.host, reply.addrlen);

	if (!(strncmp(reply.msgtype, ADDRMSG, MSGTYPESIZ)==0)) {
		if (strncmp(reply.msgtype, BADMSGMSG, MSGTYPESIZ)==0) {
			logaudit("getaddrs: a server says what\n");
			pthread_mutex_lock(&stats.st_mutex);
			stats.st_concount--;
			stats.nclientconns--;
			stats.st_contotal++;
			stats.st_contime += time(0) - start;
			pthread_mutex_unlock(&stats.st_mutex);
			return 0;
		} 
	
		if (strcmp(servmsg.filename, reply.filename)!=0) {
			logaudit("getaddrs: requested %s, server sent %s.\n", servmsg.filename, reply.filename);
			removeaddrs(reply);
			pthread_mutex_lock(&stats.st_mutex);
			stats.st_concount--;
			stats.nclientconns--;
			stats.st_contotal++;
			stats.st_contime += time(0) - start;
			pthread_mutex_unlock(&stats.st_mutex);
			return 0;	
		}
	
		if (strncmp(reply.msgtype, FILENOTFOUNDMSG, MSGTYPESIZ)==0) {
			logaudit("getaddrs: server said file %s not found, remove address and exit\n", servmsg.filename);
			removeaddrs(reply);
			pthread_mutex_lock(&stats.st_mutex);
			stats.st_concount--;
			stats.nclientconns--;
			stats.st_contotal++;
			stats.st_contime += time(0) - start;
			pthread_mutex_unlock(&stats.st_mutex);
			return 0;
		}
	} else {
		logaudit("getaddrs: server sent %d addresses\n", atoi(reply.chunkno));	
		addresses=malloc(atoi(reply.chunkno) * sizeof(addrs));
		memset(addresses, 0, atoi(reply.chunkno) * sizeof(addrs));
		//receive the data
		/* read in the data */
		bytescopied=0;
		do {
			count = read(sock, databuf, atoi(reply.chunksize));
			logaudit("getaddrs: reading data, count is: %d\n", count);
			if (count < 0) {
				logaudit("getaddrs: %s\n", strerror(errno));
				(void) pthread_mutex_lock(&stats.st_mutex);
				stats.st_concount--;
				stats.nclientconns--;
				(void) pthread_mutex_unlock(&stats.st_mutex); 
				return 0;
			}
			memcpy (&servdata[bytescopied], databuf, count);	
			bytescopied+=count;	
			(void) pthread_mutex_lock(&stats.st_mutex);
			stats.st_bytecount += count;
			(void) pthread_mutex_unlock(&stats.st_mutex); 
		} while (bytescopied < atoi(reply.chunksize));	
		
		//printf("getaddrs: for %s, bytescopied is %d, servdata is %.*s\n", servmsg.filename, bytescopied, bytescopied, servdata);
		for (i=0; i<atoi(reply.chunkno); i++) {
			logaudit("getaddrs: for %s, servdata[%d:%d] is %.*s\n", servmsg.filename, i*MAXADDRLEN, i*MAXADDRLEN+MAXADDRLEN-1, MAXADDRLEN, &servdata[i*MAXADDRLEN]); 
		}

		//parse it into addresses
		for (i=0; i<atoi(reply.chunkno); i++) {		
			memcpy(addresses[i].host, &servdata[i*MAXADDRLEN], MAXADDRLEN);	
			addresses[i].addrlen=MAXADDRLEN;
			logaudit("getaddrs: for %s, &servdata[%d] is %.*s and addresses[%d].host is %.*s\n", servmsg.filename, i*MAXADDRLEN, MAXADDRLEN, &servdata[i*MAXADDRLEN], i, MAXADDRLEN, addresses[i].host);
		}
		//send addresses list to addaddrslist
		addaddrslist (reply, addresses, atoi(reply.chunkno));
	}

	pthread_mutex_lock(&stats.st_mutex);
	stats.st_concount--;
	stats.nclientconns--;
	stats.st_contotal++;
	stats.st_contime += time(0) - start;
	pthread_mutex_unlock(&stats.st_mutex);

	return 0;
}
