/* btd.c - main, btd, prstats, more */

#include "bt.h"

/*------------------------------------------------------------------------
 * btd - Concurrent TCP server for BT service 
 *------------------------------------------------------------------------
 */
int btd (void)
{
	pthread_t	th;
	pthread_attr_t	ta;
	char	*service = BTPORT;	/* service name or port number	*/
	struct	sockaddr_in fsin;	/* the address of a client	*/
	unsigned int	alen;		/* length of client's address	*/
	int	msock;			/* master server socket		*/
	int	ssock;			/* slave server socket		*/
	char	*port;
	int *socks[QLEN];
	int counter=0;

	logaudit("btd: starting service on port %s\n", BTPORT);
	msock = passiveTCP(service, QLEN);

	if (msock<0) {
		logaudit("btd: error creating msock, exiting\n");
		return 0;
	}

	(void) pthread_attr_init(&ta);
	(void) pthread_attr_setdetachstate(&ta, PTHREAD_CREATE_DETACHED);

	for (counter=0; counter<QLEN; counter++) {
		socks[counter] = (int *) malloc(sizeof(int));
		*socks[counter] = counter;
		//printf("proof we can access *socks[%d]: %d\n", counter, *socks[counter]);
	}
	
	while (1) {
		counter=(counter+1)%QLEN;
		alen = sizeof(fsin);
		ssock = accept(msock, (struct sockaddr *)&fsin, &alen);
		//printf("accepted. ssock=%d\n", ssock);
		//printf("counter: %d\n", counter);
		//printf("*socks[counter] = %d\n", *socks[counter]);
		if (ssock < 0) {	
			if (errno == EINTR)
				continue;
			logaudit("accept(btd): %s\n", strerror(errno));
			return 0;
		}

		*socks[counter] = ssock;
		//printf("set socks[%d]\n", counter);
		if (pthread_create(&th, &ta, (void * (*)(void *))btdresp,
		    (void *)socks[counter]) < 0) {
			logaudit("pthread_create(btd): %s\n", strerror(errno));
			return 0;
		}
	}
}

/*------------------------------------------------------------------------
 * btdresp - take a bt message and send a response
 *------------------------------------------------------------------------
 */
int
btdresp (void* insock)
{
	int sock = * (int *) insock;
	time_t	start;
	char	buf[BUFSIZ];
	int	count;
	char 	clientmsg[BUFSIZ];
	int 	bytescopied=0;
	int 	filesindex;
	btintmsg 	clientbtmsg;
	memset(&clientbtmsg, 0, sizeof(btintmsg));

	start = time(0);
	(void) pthread_mutex_lock(&stats.st_mutex);
	stats.st_concount++;
	(void) pthread_mutex_unlock(&stats.st_mutex);
	
	recvheader(&clientbtmsg, sock);

	for (clientbtmsg.filesindex=0; clientbtmsg.filesindex < NUMFILES; clientbtmsg.filesindex++) {
		/*printf("filesindex is %d, files[clientbtmsg.filesindex].filename is %s.\n", clientbtmsg.filesindex, files[clientbtmsg.filesindex].filename);*/
		/*printf("strcmp(files[clientbtmsg.filesindex].filename, filename) = %d\n", strcmp(files[clientbtmsg.filesindex].filename, clientbtmsg.filename));*/
		if (strcmp(clientbtmsg.filename, files[clientbtmsg.filesindex].filename)==0) {
			logaudit("btdresp: filename %s found at index %d\n", clientbtmsg.filename, clientbtmsg.filesindex);
			break;
		}
	}

	if ((clientbtmsg.filesindex==NUMFILES) || (clientbtmsg.filesindex<NUMFILES && (files[clientbtmsg.filesindex].sessionok==TRANSFER_STOP))) {
		logaudit("btdresp: filename %s not available\n", clientbtmsg.filename);
		sendnotfound(sock, clientbtmsg);
		(void) close (sock);
		(void) pthread_mutex_lock(&stats.st_mutex);
		stats.st_contime += time(0) - start;
		stats.st_concount--;
		stats.st_contotal++;
		(void) pthread_mutex_unlock(&stats.st_mutex);
		return 0;
	} 

	if (!(strcmp(clientbtmsg.msgtype, CHUNKMSG))) { /* message is a request for data */
		//memcpy(clientbtmsg.chunkno, &clientmsg[FILENAMESIZ+MSGTYPESIZ], CHUNKNOSIZ);
		logaudit("msgtype %s, sending chunk number %.*s\n", clientbtmsg.msgtype, CHUNKNOSIZ, clientbtmsg.chunkno);	
		sendchunk (sock, clientbtmsg);
	} else if (!(strcmp(clientbtmsg.msgtype, ADDRMSG))) { /* message is a request for addresses */
		logaudit("msgtype %s, sending addresses\n", clientbtmsg.msgtype);
		sendaddrs (sock, clientbtmsg);
	} else { /* message type not found */
		logaudit("msgtype %s, not found\n", clientbtmsg.msgtype);
		sendbadmsg (sock, clientbtmsg);
		logerror (sock, clientmsg, bytescopied);
	}
	
	(void) shutdown(sock, 0);
	(void) pthread_mutex_lock(&stats.st_mutex);
	stats.st_contime += time(0) - start;
	stats.st_concount--;
	stats.st_contotal++;
	(void) pthread_mutex_unlock(&stats.st_mutex);
	return 0;
}

/*------------------------------------------------------------------------
 * sendchunk - send a chunk of data
 *------------------------------------------------------------------------
 */
int sendchunk (int sock, btintmsg clientmsg) {
	btmsg msg;
	size_t datasize;
	FILE* fp;
	long offset;
	size_t bytesread;
	char* buf[files[clientmsg.filesindex].chunksize];
	int i;
	div_t q;
	int chunk2send;

	chunk2send=atoi(clientmsg.chunkno);
	logaudit("sendchunk: chunk2send (atoi(clientmsg.chunkno)) is %d, string of clientmsg.chunkno is %.*s\n", chunk2send, CHUNKNOSIZ, clientmsg.chunkno);

	/* check here if we have that chunk*/
	if (chunk2send>(files[clientmsg.filesindex].nchunks-1)) {
		/*client asked for a chunk greater than the number in this file, set chunk to something reasonable */
		chunk2send=0;
		logaudit("sendchunk: client asked for chunk %d, but there are only %d in file %s, setting to 0\n", atoi(clientmsg.chunkno), files[clientmsg.filesindex].nchunks, files[clientmsg.filesindex].filename);
	}
	logaudit("sendchunk: files[%d].chunks[%d].want==%d, have==%d\n", clientmsg.filesindex, chunk2send, files[clientmsg.filesindex].chunks[chunk2send].want, files[clientmsg.filesindex].chunks[chunk2send].have);

	/* do we have it or not?*/
	i=0;
	while ((files[clientmsg.filesindex].chunks[chunk2send].want==1) && (i<files[clientmsg.filesindex].nchunks)) {
		logaudit("sendchunk: files[%d].chunks[%d].want==%d\n", clientmsg.filesindex, chunk2send, files[clientmsg.filesindex].chunks[chunk2send].want);
		if (chunk2send>files[clientmsg.filesindex].nchunks) {
			chunk2send=0;
		} else {
			chunk2send++;
		}
		i++;
	}

	if (i==files[clientmsg.filesindex].nchunks) {
		logaudit("sendchunk: no chunks for this file, sending not found message\n");
		sendnotfound(sock, clientmsg);
		return 0;
	}

	logaudit("sendchunk: request for chunk %d, decided to send chunk %d\n", atoi(clientmsg.chunkno), chunk2send);

	updateaddrs(clientmsg, sock);

	offset=chunk2send*files[clientmsg.filesindex].chunksize;
	logaudit("sendchunk: atoi(clientmsg.chunkno) = %d\n", atoi(clientmsg.chunkno));
	logaudit("sendchunk: request for chunk %d, sending chunk %d, making offset %d*%d=%d\n", atoi(clientmsg.chunkno), chunk2send, chunk2send, files[clientmsg.filesindex].chunksize, offset);

	datasize=files[clientmsg.filesindex].chunksize;
	/* check here if we want to read the last chunk, which will be smaller*/

	if (chunk2send==files[clientmsg.filesindex].nchunks-1) {
		q=div(files[clientmsg.filesindex].filesize,files[clientmsg.filesindex].chunksize);
		logaudit("sendchunk: datasize for the end chunk (chunk %d) is actually %d, not %d\n", chunk2send, q.rem, files[clientmsg.filesindex].chunksize);
		datasize=q.rem;
	}

	//pthread_mutex_lock(&files[clientmsg.filesindex].filedesc_mutex);
	logaudit("sendchunk: filename to open is %s\n", clientmsg.filename);
	fp=fopen(clientmsg.filename, "r");
	fseek(fp, offset, SEEK_SET);
	if((bytesread=fread(buf, 1, datasize, fp))<=0) {
		logaudit ("sendchunk: %s\n", strerror(errno));
		return 0;
	}

	logaudit("sendchunk: read %d bytes from %s\n", bytesread, clientmsg.filename);
	fclose(fp);
	//pthread_mutex_unlock(&files[clientmsg.filesindex].filedesc_mutex);

	memset (&msg, 0, sizeof msg);
	
	memcpy(msg.filename, clientmsg.filename, FILENAMESIZ);
	memcpy(msg.msgtype, clientmsg.msgtype, MSGTYPESIZ);
	snprintf(msg.chunkno, CHUNKNOSIZ, "%d", chunk2send);
	snprintf(msg.chunksize, CHUNKLENGTHSIZ, "%d", datasize);
	/*memcpy(msg.chunkno, &chunk2send, CHUNKNOSIZ);
	memcpy(msg.chunksize, &datasize, CHUNKLENGTHSIZ);*/

	logaudit("sendchunk: copying %d bytes to msg.data\n", bytesread);
	memcpy(msg.data, buf, bytesread);

	if (write(sock, &msg, sizeof msg) < 0) {
		logaudit ("sendchunk: %s\n", strerror(errno));
		return 0;
	}

	logaudit("sendchunk: successfully sent msgtype %.*s\n", MSGTYPESIZ, msg.msgtype);
	return 0;
}

/*------------------------------------------------------------------------
 * sendaddrs - send an address list
 *------------------------------------------------------------------------
 */
int sendaddrs (int sock, btintmsg clientmsg) {
	btmsg msg;
	int naddrsused=files[clientmsg.filesindex].naddrsused;
	int addrsonlist=0;
	int datasize;
	int i;
	struct sockaddr_in cli_addr;
	socklen_t sin_len;

	sin_len = sizeof (cli_addr);
	getpeername (sock, (struct sockaddr *) &cli_addr, &sin_len);

	//char* addresses[MAXADDRLEN*naddrsused];
	//memset(addresses, 0, MAXADDRLEN*naddrsused);
	datasize=0;

	memset (&msg, 0, sizeof msg);
	memcpy (msg.filename, clientmsg.filename, FILENAMESIZ);
	memcpy (msg.msgtype, clientmsg.msgtype, MSGTYPESIZ);

	pthread_mutex_lock(&files[clientmsg.filesindex].addresses_mutex);

	for(i=0;i<naddrsused; i++) {
		//don't send more addresses than will fit in one message
		if (addrsonlist*MAXADDRLEN+MAXADDRLEN > CHUNKSIZE) {
			break;
		} else if (!(strcmp(files[clientmsg.filesindex].addresses[i].host, inet_ntoa(cli_addr.sin_addr)))) {
			logaudit("sendaddrs: files[clientmsg.filesindex].addresses[%d].host is %s, client address is %s, keeping client's own address out of message\n", i, files[clientmsg.filesindex].addresses[i].host, inet_ntoa(cli_addr.sin_addr));
			continue;
		} else {
			/*add the address string to msg.data*/
			logaudit("sendaddrs: files[clientmsg.filesindex].addresses[%d].host is %s, addrlen is %d\n", i, files[clientmsg.filesindex].addresses[i].host, files[clientmsg.filesindex].addresses[i].addrlen);
			memcpy(&msg.data[addrsonlist*MAXADDRLEN], files[clientmsg.filesindex].addresses[i].host, files[clientmsg.filesindex].addresses[i].addrlen);
			logaudit("sendaddrs: &msg.data[%d] is %.*s\n", addrsonlist*MAXADDRLEN, MAXADDRLEN, &msg.data[addrsonlist*MAXADDRLEN]);
			datasize+=files[clientmsg.filesindex].addresses[i].addrlen;
			addrsonlist++;
		}
	}

	pthread_mutex_unlock(&files[clientmsg.filesindex].addresses_mutex);

	logaudit("sendaddrs: client message:\n clientmsg.filename: %s\n clientmsg.msgtype: %s\n clientmsg.chunkno: %s\n clientmsg.chunksize: %s\n", clientmsg.filename, clientmsg.msgtype, clientmsg.chunkno, clientmsg.chunksize);

	sprintf (msg.chunkno, "%d", addrsonlist);
	logaudit("chunkno (number of addresses) is %.*s\n", CHUNKNOSIZ, msg.chunkno);
	sprintf (msg.chunksize, "%d", datasize);
	//printf("chunksize is %.8s\n", msg.chunksize);
	/*memcpy (msg.chunksize, &datasize, CHUNKLENGTHSIZ);*/

	logaudit("sendaddrs: message to send:\n msg.filename: %.*s\n msg.msgtype: %.*s\n msg.chunkno: %.*s\n msg.chunksize: %.*s\n", FILENAMESIZ, msg.filename, MSGTYPESIZ, msg.msgtype, CHUNKNOSIZ, msg.chunkno, CHUNKLENGTHSIZ, msg.chunksize);

	if (write(sock, &msg, sizeof msg) < 0) {
		printf ("bt sendaddrs: %s\n", strerror(errno));
		return 0;
	}
	
	logaudit("sendaddrs: successfully sent msgtype %s\n", msg.msgtype);

	return 0;
}

/*------------------------------------------------------------------------
 * sendbadmsg - send a message indicating that the message type was not recognized
 *------------------------------------------------------------------------
 */
int sendbadmsg (int sock, btintmsg clientmsg) {
	btmsg msg;
	memset (&msg, 0, sizeof msg);
	memcpy (msg.filename, files[clientmsg.filesindex].filename, FILENAMESIZ);
	memcpy (msg.msgtype, BADMSGMSG, MSGTYPESIZ);
	if (write(sock, &msg, sizeof msg) < 0) {
		logaudit("sendbadmsg: %s\n", strerror(errno));
		return 0;
	}
	logaudit("sendbadmsg: successfully sent msgtype %s\n", msg.msgtype);
	return 0;
}

/*------------------------------------------------------------------------
 * sendnotfound - send a message indicating that we don't have this file
 *------------------------------------------------------------------------
 */
int sendnotfound (int sock, btintmsg clientmsg) {
	btmsg msg;
	memset (&msg, 0, sizeof msg); 
	memcpy (msg.filename, clientmsg.filename, FILENAMESIZ);
	memcpy (msg.msgtype, FILENOTFOUNDMSG, MSGTYPESIZ);
	if (write(sock, &msg, sizeof msg) <0) {
		logaudit("sendnotfound: %s\n", strerror(errno));
		return 0;
	}
	logaudit("sendnotfound: successfully sent msgtype %s\n", msg.msgtype);
	return 0;
}


