/* updatefiles and loadstate: functions for creating and otherwise messing with the global files object files */

#include "bt.h"

/*------------------------------------------------------------------------
 * loadstate - load the files list from disk, from location defined in constant "filestate"
 *------------------------------------------------------------------------
 */

int loadstate(void) {

	FILE* fp;
	int i;
	int j;
	int k;
	int bytesread=0;
	/*for each file in the files list
	//load each attribute in order except for mutexes
	//for each element on each list*/

	logaudit("loadstate: opening %s\n", &filestate);
	(void) pthread_mutex_lock(&filestate_mutex);
	fp=fopen(filestate, "r");
	logaudit("loadstate: nfiles is %d\n", nfiles);	
	bytesread+=fscanf(fp, "%d", &nfiles);
	logaudit("loadstate: nfiles is now %d\n", nfiles);	
	for (i=0; i<nfiles; i++) {

		if (pthread_mutex_lock(&files[i].filedesc_mutex)<0) {
			if (errno==EINVAL) { 
				logaudit("files[%d].filedesc_mutex was not initialized, initializing and locking\n", i);
				pthread_mutex_init(&files[i].filedesc_mutex, 0);
				(void) pthread_mutex_lock(&files[i].filedesc_mutex);
			} else {
				errexit("loadstate: %s\n", strerror(errno));
			}
		}
		/*files[i] is definitely already allocated to MAXFILES*/

		bytesread+=fscanf(fp, "%d", &files[i].filesize);

		//get the newline after filesize, put it in the buffer that is about to be overwritten
		fgets(files[i].filename, FILENAMESIZ+1, fp);
	
		if (fgets(files[i].filename, FILENAMESIZ+1, fp)==NULL) {
			logaudit("error fgets: no filename found");
			continue;
		}
		logaudit("files[%d].filename is: %s\n", i, files[i].filename);
		//remove the newline that fgets appended to files[i].filename
		files[i].filename[strlen(files[i].filename)-1]=0;
		logaudit("files[%d].filename is now: %s\n", i, files[i].filename);
		bytesread+=fscanf(fp, "%d", &files[i].transferok);
		files[i].sessionok=files[i].transferok;
		bytesread+=fscanf(fp, "%d", &files[i].chunksize);
		bytesread+=fscanf(fp, "%d", &files[i].nchunks);

		logaudit("i is %d, filesize is %d, filename is %s, transferok is %d, chunksize is %d, nchunks is %d\n", i, files[i].filesize, files[i].filename, files[i].transferok, files[i].chunksize, files[i].nchunks);

		/*the number of chunks of the file shouldn't change*/
		/*this line is only necessary if files[i].chunks doesn't point anywhere yet*/
		files[i].chunks=realloc(files[i].chunks, files[i].nchunks * sizeof(chunk));

		for (j=0; j<files[i].nchunks; j++) {
			bytesread+=fscanf(fp, "%d", &files[i].chunks[j].chunkno);
			bytesread+=fscanf(fp, "%d", &files[i].chunks[j].checksum);
			bytesread+=fscanf(fp, "%d", &files[i].chunks[j].want);
			bytesread+=fscanf(fp, "%d", &files[i].chunks[j].have);	
			//printf("%s: chunkno is %d, checksum is %d, want:%d, have:%d\n", files[i].filename, files[i].chunks[j].chunkno, files[i].chunks[j].checksum, files[i].chunks[j].want, files[i].chunks[j].have);
			//printf("j is %d\n", j);
		}
		
		logaudit("about to scan naddrs\n");
		bytesread+=fscanf(fp, "%d", &files[i].naddrs);	
		logaudit("about to scan naddrsused\n");
		bytesread+=fscanf(fp, "%d", &files[i].naddrsused);	
		logaudit("%s: naddrs is %d, naddrsused is %d\n", files[i].filename, files[i].naddrs, files[i].naddrsused);
		files[i].filesindex=i;
		(void) pthread_mutex_unlock(&files[i].filedesc_mutex);

		if (pthread_mutex_lock(&files[i].addresses_mutex)<0) {
			if (errno==EINVAL) { 
				logaudit("files[%d].addresses_mutex was not initialized, initializing and locking\n", i);
				pthread_mutex_init(&files[i].addresses_mutex, 0);
				(void) pthread_mutex_lock(&files[i].addresses_mutex);
			} else {
				errexit("loadstate addresses_mutex: %s\n", strerror(errno));
			}
		}
		/*again, this is only required if files[i].addresses doesn't point anywhere yet*/
		files[i].addresses=realloc(files[i].addresses, files[i].naddrs * sizeof(addrs));
		for (k=0; k<files[i].naddrsused; k++) {
			bytesread+=fscanf(fp, "%d", &files[i].addresses[k].addrlen);
			bytesread+=fscanf(fp, "%s", files[i].addresses[k].host);
			logaudit("%s: addrlen is %d, host is %s\n", files[i].filename, files[i].addresses[k].addrlen, files[i].addresses[k].host);
		}
		(void) pthread_mutex_unlock(&files[i].addresses_mutex);

		/*there's an extra linebreak between files*/
		bytesread+=fscanf(fp, "\n");			
	}
		
	fclose(fp);
	(void) pthread_mutex_unlock(&filestate_mutex);
	return 0;
}

/*------------------------------------------------------------------------
 * updatefiles - create the file list at startup, or whenever
 * - sets want/have by checking if the referenced file exists
 *------------------------------------------------------------------------
 */

int updatefiles(void) {
	
	int i;
	char currline[CHUNKNOSIZ+1+CHUNKLENGTHSIZ];
	int chunkindex;
	struct dirent **metafiles;
	int n;
	char fname[BTNAME_MAX];
	int chunksize;
	int nchunks;
	FILE* fp;
	char tempfname[BTNAME_MAX];
	int want;
	int have;

	char* ext_start;

	logaudit("updating files\n");
 	/*files = malloc(NUMFILES * sizeof(filedesc));
	memset(files, 0, NUMFILES * sizeof(filedesc));*/

	n = scandir(".", &metafiles, 0, 0);
	/*printf("n is %d\n", n);*/
	(void) pthread_mutex_lock(&filestate_mutex);
	for (i=0; i<n; i++) {
		if (nfiles>=NUMFILES) {
			break;
		}
		/*printf("checking metafiles[%d]->d_name, %s\n", i, metafiles[i]->d_name);*/

                ext_start = strstr (metafiles[i]->d_name, METAFILEEXT);
                /*printf ("After strstr. ext_start is NULL? %s\n", ext_start ? "No" : "Yes");*/
		if ((ext_start != NULL) && (strcmp(ext_start, METAFILEEXT)==0)) {
			logaudit("found a metadata file called %s\n", metafiles[i]->d_name);
			strncpy(fname, metafiles[i]->d_name, NAME_MAX);
			loadmetafile(fname);
		}
		/*printf("if statement over, i=%d\n", i);*/
	}
	(void) pthread_mutex_unlock(&filestate_mutex);
	logaudit("in updatefiles, saving state to disk\n");
	savestate();
	logaudit("done saving state\n");
	return 0;
}

/*------------------------------------------------------------------------
 * infiletable - returns 1 if the filename referenced by this metafile already has an entry in the filetable, otherwise 0
 *------------------------------------------------------------------------
 */

int infiletable (char *filename) {

	int i;
	logaudit("infiletable: filename is %s\n", filename);
	for (i=0; i<nfiles; i++) {
		logaudit("infiletable: files[%d].filename is %s\n", i, files[i].filename);
		if (strcmp(files[i].filename, filename)==0) {
			return 1;
		}
	}

	return 0;
}

/*------------------------------------------------------------------------
 * whereinfiletable - returns the location of an entry in the filetable for filename, -1 if not found 
 *------------------------------------------------------------------------
 */

int whereinfiletable (char *filename) {

	int i;
	logaudit("whereinfiletable: filename is %s\n", filename);
	for (i=0; i<nfiles; i++) {
		logaudit("whereinfiletable: files[%d].filename is %s\n", i, files[i].filename);
		if (strcmp(files[i].filename, filename)==0) {
			return i;
		}
	}

	return -1;
}

/*------------------------------------------------------------------------
 * loadmetafile - loads information from a single metafile into the file table
 *------------------------------------------------------------------------
 */

int loadmetafile (char *fname) {

	int chunkindex;
	int n;
	int chunksize;
	int nchunks;
	FILE* fp;
	char tempfname[BTNAME_MAX];
	int want;
	int have;
	int index=nfiles;

	fp = fopen(fname, "r");

	if (pthread_mutex_lock(&files[index].filedesc_mutex)<0) {
		if (errno==EINVAL) { 
			logaudit("files[%d].filedesc_mutex was not initialized, initializing and locking\n", index);
			pthread_mutex_init(&files[index].filedesc_mutex, 0);
			(void) pthread_mutex_lock(&files[index].filedesc_mutex);
		} else {
			logaudit("loadmetafile: %s\n", strerror(errno));
			return -1;
		}			
	}

	(void) pthread_mutex_unlock(&files[index].filedesc_mutex);

	if (pthread_mutex_lock(&files[index].addresses_mutex)<0) {
		if (errno==EINVAL) { 
			logaudit("files[%d].addresses_mutex was not initialized, initializing and locking\n", index);
			pthread_mutex_init(&files[index].addresses_mutex, 0);
			(void) pthread_mutex_lock(&files[index].addresses_mutex);
		} else {
			logaudit("loadmetafile: %s\n", strerror(errno));
			return -1;
		}
	}

	(void) pthread_mutex_unlock(&files[index].addresses_mutex);

	if (fgets(tempfname, BTNAME_MAX+1, fp)==NULL) {
		logaudit("error fgets: no filename found");
		return -1;
	} else {
		logaudit("tempfname is: %s\n", tempfname);
		//remove the newline that fgets appended to tempfname
		tempfname[strlen(tempfname)-1]=0;
		logaudit("tempfname is now: %s\n", tempfname);
		if (infiletable(tempfname)) {
			if (files[whereinfiletable(tempfname)].transferok==TRANSFER_STOP) {
				logaudit("%s previously cancelled, skipping\n", tempfname);
				return 1;
			} else {
				logaudit("%s already loaded, skipping\n", tempfname);
				return 1;
			}
		} else {
			memcpy(files[index].filename, tempfname, strlen(tempfname));
			//sscanf(tempfname, "%s", files[index].filename);
		}
		logaudit("files[%d].filename is %s\n", index, files[index].filename);
	}

	if (access(files[index].filename, F_OK)<0) {
		logaudit("access: %s\n", strerror(errno));
		logaudit("no file called %s found. want=1, have=0.\n", files[index].filename);
		want=1;
		have=0;
	} else {
		logaudit("file called %s found. want=0, have=1.\n", files[index].filename);
		want=0;
		have=1;
	}

	if (fscanf(fp, "%d", &files[index].filesize)<1) {
		logaudit("error fscanf: no filesize found");
		return -1;
	} else {
		logaudit("files[%d].filesize is %d\n", index, files[index].filesize);
	}

	files[index].naddrs = 10; /* let's start with allocating 10 addresses */
	files[index].naddrsused = 1; /*we expect 1 address in each metafile */
	files[index].addresses = malloc(files[index].naddrs * sizeof(addrs));

	if (fscanf(fp, "%s", &files[index].addresses[0].host)<1) {
		logaudit("error reading address from %s\n", fname);
		return -1;
	} else {
		files[index].addresses[0].addrlen = MAXADDRLEN;
		logaudit("files[%d].addresses->host is %s.\n", index, files[index].addresses[0].host);
		logaudit("files[%d].addresses->addrlen is %d\n", index, files[index].addresses[0].addrlen);
	}

	if (fscanf(fp, "%d", &files[index].chunksize)<1) {
		logaudit("error reading chunksize from %s\n", fname);
		return -1;
	} else {
		logaudit("files[%d].chunksize is %d\n", index, files[index].chunksize);
	}

	if (fscanf(fp, "%d", &nchunks)<1) {
		logaudit("error reading nchunks from %s\n", fname);
		return -1;
	} else {
		files[index].chunks = malloc(nchunks * sizeof(chunk));
		files[index].nchunks = nchunks;
		logaudit("nchunks is %d\n", nchunks);
	}

	chunkindex=0;
	while (chunkindex<nchunks) {
		fscanf(fp, "%d", &files[index].chunks[chunkindex].chunkno);
		//printf("chunkno is %d\n", files[index].chunks[chunkindex].chunkno);
		fscanf(fp, "%d", &files[index].chunks[chunkindex].checksum);
		//printf("checksum is %d\n", files[index].chunks[chunkindex].checksum);
		files[index].chunks[chunkindex].want=want;
		files[index].chunks[chunkindex].have=have;
		chunkindex++;			
	}
	fclose(fp);
	files[index].filesindex=index;
	files[index].transferok=TRANSFER_OK;
	files[index].sessionok=files[index].transferok;

	//Ensure that nfiles is not incremented needlessly.
	if (index==nfiles) {
		nfiles++;
	}
	
	return 0;
}
