/* savestate: saves the current state of the global file object files to the local file named in the constant "filestate" */

#include "bt.h"

/*------------------------------------------------------------------------
 * savestate - save the files list to disk, every time it changes
 *------------------------------------------------------------------------
 */

int savestate(void) {

	FILE* fp;
	int i;
	int j;
	int k;
	int byteswritten=0;
	/*for each file in the files list
	//save each attribute in order except for mutexes
	//for each element on each list*/
	logaudit("savestate: about to open the filestate file\n");
	(void) pthread_mutex_lock(&filestate_mutex);
	fp=fopen(filestate, "w");
	byteswritten+=fprintf(fp, "%d\n", nfiles);
	for (i=0; i<nfiles; i++) {
		(void) pthread_mutex_lock(&files[i].filedesc_mutex);
		//printf("savestate: locked files[%d].filedesc_mutex\n", i);
		byteswritten+=fprintf(fp, "%d\n", files[i].filesize);	
		byteswritten+=fprintf(fp, "%s\n", files[i].filename);	
		byteswritten+=fprintf(fp, "%d\n", files[i].transferok);
		byteswritten+=fprintf(fp, "%d\n", files[i].chunksize);
		byteswritten+=fprintf(fp, "%d\n", files[i].nchunks);	

		for (j=0; j<files[i].nchunks; j++) {
			byteswritten+=fprintf(fp, "%d %d %d %d\n", files[i].chunks[j].chunkno, files[i].chunks[j].checksum, files[i].chunks[j].want, files[i].chunks[j].have);		
		}
					
		byteswritten+=fprintf(fp, "%d\n", files[i].naddrs);	
		byteswritten+=fprintf(fp, "%d\n", files[i].naddrsused);	
		(void) pthread_mutex_unlock(&files[i].filedesc_mutex);
		//printf("savestate: unlocked files[%d].filedesc_mutex\n", i);
		(void) pthread_mutex_lock(&files[i].addresses_mutex);
		//printf("savestate: locked files[%d].addresses_mutex\n", i);
		for (k=0; k<files[i].naddrsused; k++) {
			byteswritten+=fprintf(fp, "%d %.*s\n", files[i].addresses[k].addrlen, files[i].addresses[k].addrlen, files[i].addresses[k].host);
		}
		(void) pthread_mutex_unlock(&files[i].addresses_mutex);
		//printf("savestate: unlocked files[%d].addresses_mutex\n", i);
		/*put an extra linebreak between files*/
		byteswritten+=fprintf(fp, "\n");			
	}
	logaudit("savestate: about to close the filestate file\n");	
	fclose(fp);
	(void) pthread_mutex_unlock(&filestate_mutex);
	return 0;
}

