/* logerror and prstats - maintenance and upkeep functions */

#include "bt.h"

/*------------------------------------------------------------------------
 * logerror - log an error
 *------------------------------------------------------------------------
 */
int logerror (int sock, char* request, int length) {

	time_t now;
	int count;
	FILE* fp = fopen(errorlog, "a");

	now = time(0);
	fprintf(fp, "%s --- request: ", ctime(&now));
	count=fwrite(request, length, 1, fp);
	fclose(fp);
	printf("logerror: successfully logged error\n");
	return 0;
}

/*------------------------------------------------------------------------
 * logaudit - log all activity messages to auditlog
 *------------------------------------------------------------------------
 */
int logaudit(const char *format, ...)
{
	time_t now;
	va_list	args;

	FILE* fp=fopen(auditlog, "a");
	now=time(0);
	fprintf(fp, "%s --- ", ctime(&now));
	va_start(args, format);
	vfprintf(fp, format, args);
	va_end(args);
	fclose(fp);
}


/*------------------------------------------------------------------------
 * prstats - print server statistical data
 * Adapted from Comer.
 *------------------------------------------------------------------------
 */
void
prstats(void)
{
	time_t	now;

	//while (1) {
		//(void) sleep(INTERVAL);

		(void) pthread_mutex_lock(&stats.st_mutex);
		now = time(0);
		(void) printf("--- %s", ctime(&now));
		(void) printf("%-32s: %u\n", "Current connections",
			stats.st_concount);
		(void) printf("%-32s: %u\n", "Completed connections",
			stats.st_contotal);
		if (stats.st_contotal) {
			(void) printf("%-32s: %.2f (secs)\n",
				"Average complete connection time",
				(float)stats.st_contime /
				(float)stats.st_contotal);
			(void) printf("%-32s: %.2f\n",
				"Average byte count",
				(float)stats.st_bytecount /
				(float)(stats.st_contotal +
				stats.st_concount));
		}
		(void) printf("%-32s: %lu\n\n", "Total byte count",
			stats.st_bytecount);
		(void) pthread_mutex_unlock(&stats.st_mutex);

	//}
}

/*------------------------------------------------------------------------
 * prfiles - prints all files that bt knows about 
 *------------------------------------------------------------------------
 */

int prfiles (void) {

	int i;
	int j;
	int nhave;
	printf("\t# filename\thave/total\tpaused\ttransfer ok\n");
	for (i=0; i<nfiles; i++) {
		if (files[i].transferok==TRANSFER_STOP) {
			continue;
		}
		nhave=0;
		for (j=0; j<files[i].nchunks; j++) {
			nhave+=files[i].chunks[j].have;
		}
		printf("\t%d %s\t%d/%d\t%d\t%d\n", i, files[i].filename, nhave, files[i].nchunks, !files[i].sessionok, files[i].transferok);
	}

	return 0;
}
